import sys
import pytest
import subprocess
from unittest import mock
from unittest.mock import MagicMock, patch
import torch

from mm_poe.cli import main


@patch("mm_poe.cli.set_seed")
@patch("mm_poe.cli.load_model")
@patch("mm_poe.cli.download_model")
@patch("mm_poe.cli.questionary.select")
@patch("mm_poe.cli.questionary.path")
@patch("mm_poe.cli.questionary.text")
def test_main(
    mock_text,
    mock_path,
    mock_select,
    mock_download_model,
    mock_load_model,
    mock_set_seed,
):
    # Mock the inputs provided by questionary
    mock_select.return_value.ask.side_effect = [
        "GIT",  # args.model_family
        "microsoft/git-base-vqav2",  # args.checkpoint
        "FP32",  # args.loading_precision
        "language_modeling",  # args.scoring_method_for_process_of_elimination
        "below_average",  # args.mask_strategy_for_process_of_elimination
        "0",  # args.label
    ]

    mock_path.return_value.ask.side_effect = [
        "./models/",  # args.output_dir
        "./images/image.png",  # args.image_path
    ]

    mock_text.return_value.ask.side_effect = [
        "What is in the image?",  # args.question
        "cat,dog,horse",  # args.choices
    ]

    # Mock the download_model to prevent actual execution
    mock_download_model.return_value = 0

    # Mock the load_model function to return mock model and tokenizer
    mock_model = MagicMock()
    mock_tokenizer = MagicMock()
    mock_load_model.return_value = (mock_model, mock_tokenizer)

    # Mock the device
    with patch("torch.device") as mock_device:
        mock_device.return_value = "cpu"

        # Mock other functions called within main
        with patch("mm_poe.cli.load_data") as mock_load_data, patch(
            "torch.utils.data.DataLoader"
        ) as mock_data_loader_class, patch(
            "mm_poe.cli.inference_language_modeling"
        ) as mock_inference_lm, patch(
            "mm_poe.cli.inference_process_of_elimination"
        ) as mock_inference_poe, patch(
            "mm_poe.cli.compute_mask_process_of_elimination"
        ) as mock_compute_mask, patch(
            "mm_poe.cli.create_multiple_choice_prompt"
        ) as mock_create_mcp:

            # Mock the datasets returned by load_data
            mock_dataset = MagicMock()
            mock_dataset.map.return_value = mock_dataset  # For the map calls
            mock_load_data.return_value = (
                ["hypothesis0", "hypothesis1", "hypothesis2"],  # ending_names
                "premise",  # header_name
                "image_path",  # image_header_name
                mock_dataset,  # raw_dataset
                mock_dataset,  # n_shot_dataset
            )

            # Mock the DataLoader
            mock_data_loader = MagicMock()
            mock_data_loader_class.return_value = mock_data_loader

            # Mock inference functions
            # For scoring_method == 'language_modeling'
            mock_inference_lm.return_value = (
                torch.tensor([[0.1, 0.2, 0.7]]),
                None,
                None,
                torch.tensor([2]),
            )
            # For inference_process_of_elimination
            mock_inference_poe.return_value = (
                torch.tensor([[0.1, 0.2, 0.7]]),
                1.0,
                None,
                torch.tensor([2]),
            )

            # Mock compute_mask_process_of_elimination
            mock_compute_mask.return_value = torch.tensor([[0, 1, 1]])

            # Mock create_multiple_choice_prompt
            def mock_create_mcp_fn(example, **kwargs):
                return example

            mock_create_mcp.side_effect = mock_create_mcp_fn

            # Run the main function
            main()

            # Assertions to check if functions were called as expected
            mock_set_seed.assert_called_once_with(0)
            mock_download_model.assert_called()
            mock_load_model.assert_called()
            mock_load_data.assert_called()
            mock_inference_lm.assert_called()
            mock_inference_poe.assert_called()


@patch("mm_poe.cli.set_seed")
@patch("mm_poe.cli.load_model")
@patch("mm_poe.cli.download_model")
@patch("mm_poe.cli.questionary.select")
@patch("mm_poe.cli.questionary.path")
@patch("mm_poe.cli.questionary.text")
def test_main_with_calibration_lowest(
    mock_text,
    mock_path,
    mock_select,
    mock_download_model,
    mock_load_model,
    mock_set_seed,
):
    mock_select.return_value.ask.side_effect = [
        "BLIP2-OPT",  # args.model_family
        "Salesforce/blip2-opt-2.7b",  # args.checkpoint
        "FP16",  # args.loading_precision
        "calibration",  # args.scoring_method_for_process_of_elimination
        "lowest",  # args.mask_strategy_for_process_of_elimination
        "1",  # args.label
    ]

    mock_path.return_value.ask.side_effect = [
        "./models/",  # args.output_dir
        "./images/image.png",  # args.image_path
    ]

    mock_text.return_value.ask.side_effect = [
        "Describe the image.",  # args.question
        "apple,banana,orange",  # args.choices
    ]

    mock_download_model.return_value = 0
    mock_model = MagicMock()
    mock_tokenizer = MagicMock()
    mock_load_model.return_value = (mock_model, mock_tokenizer)

    with patch("torch.device") as mock_device:
        mock_device.return_value = "cuda:0"

        with patch("mm_poe.cli.load_data") as mock_load_data, patch(
            "torch.utils.data.DataLoader"
        ) as mock_data_loader_class, patch(
            "mm_poe.cli.inference_calibration"
        ) as mock_inference_calibration, patch(
            "mm_poe.cli.inference_process_of_elimination"
        ) as mock_inference_poe, patch(
            "mm_poe.cli.compute_mask_process_of_elimination"
        ) as mock_compute_mask, patch(
            "mm_poe.cli.create_multiple_choice_prompt"
        ) as mock_create_mcp:

            mock_dataset = MagicMock()
            mock_dataset.map.return_value = mock_dataset
            mock_load_data.return_value = (
                ["hypothesis0", "hypothesis1", "hypothesis2"],
                "premise",
                "image_path",
                mock_dataset,
                mock_dataset,
            )

            mock_data_loader = MagicMock()
            mock_data_loader_class.return_value = mock_data_loader

            mock_inference_calibration.return_value = (
                torch.tensor([[0.3, 0.4, 0.3]]),
                None,
                None,
                torch.tensor([1]),
            )
            mock_inference_poe.return_value = (
                torch.tensor([[0.3, 0.4, 0.3]]),
                1.0,
                None,
                torch.tensor([1]),
            )

            mock_compute_mask.return_value = torch.tensor([[1, 0, 1]])

            def mock_create_mcp_fn(example, **kwargs):
                return example

            mock_create_mcp.side_effect = mock_create_mcp_fn

            main()

            mock_set_seed.assert_called_once_with(0)
            mock_download_model.assert_called()
            mock_load_model.assert_called()
            mock_load_data.assert_called()
            mock_inference_calibration.assert_called()
            mock_inference_poe.assert_called()


@patch("mm_poe.cli.set_seed")
@patch("mm_poe.cli.load_model")
@patch("mm_poe.cli.download_model")
@patch("mm_poe.cli.questionary.select")
@patch("mm_poe.cli.questionary.path")
@patch("mm_poe.cli.questionary.text")
def test_main_with_mcp_lowest(
    mock_text,
    mock_path,
    mock_select,
    mock_download_model,
    mock_load_model,
    mock_set_seed,
):
    mock_select.return_value.ask.side_effect = [
        "BLIP2-OPT",  # args.model_family
        "Salesforce/blip2-opt-2.7b",  # args.checkpoint
        "FP16",  # args.loading_precision
        "multiple_choice_prompt",  # args.scoring_method_for_process_of_elimination
        "lowest",  # args.mask_strategy_for_process_of_elimination
        "1",  # args.label
    ]

    mock_path.return_value.ask.side_effect = [
        "./models/",  # args.output_dir
        "./images/image.png",  # args.image_path
    ]

    mock_text.return_value.ask.side_effect = [
        "Describe the image.",  # args.question
        "apple,banana,orange",  # args.choices
    ]

    mock_download_model.return_value = 0
    mock_model = MagicMock()
    mock_tokenizer = MagicMock()
    mock_load_model.return_value = (mock_model, mock_tokenizer)

    with patch("torch.device") as mock_device:
        mock_device.return_value = "cuda:0"

        with patch("mm_poe.cli.load_data") as mock_load_data, patch(
            "torch.utils.data.DataLoader"
        ) as mock_data_loader_class, patch(
            "mm_poe.cli.inference_language_modeling"
        ) as mock_inference_language_modeling, patch(
            "mm_poe.cli.inference_process_of_elimination"
        ) as mock_inference_poe, patch(
            "mm_poe.cli.compute_mask_process_of_elimination"
        ) as mock_compute_mask, patch(
            "mm_poe.cli.create_multiple_choice_prompt"
        ) as mock_create_mcp:

            mock_dataset = MagicMock()
            mock_dataset.map.return_value = mock_dataset
            mock_load_data.return_value = (
                ["hypothesis0", "hypothesis1", "hypothesis2"],
                "premise",
                "image_path",
                mock_dataset,
                mock_dataset,
            )

            mock_data_loader = MagicMock()
            mock_data_loader_class.return_value = mock_data_loader

            mock_inference_language_modeling.return_value = (
                torch.tensor([[0.3, 0.4, 0.3]]),
                None,
                None,
                torch.tensor([1]),
            )
            mock_inference_poe.return_value = (
                torch.tensor([[0.3, 0.4, 0.3]]),
                1.0,
                None,
                torch.tensor([1]),
            )

            mock_compute_mask.return_value = torch.tensor([[1, 0, 1]])

            def mock_create_mcp_fn(example, **kwargs):
                return example

            mock_create_mcp.side_effect = mock_create_mcp_fn

            main()

            mock_set_seed.assert_called_once_with(0)
            mock_download_model.assert_called()
            mock_load_model.assert_called()
            mock_load_data.assert_called()
            mock_inference_language_modeling.assert_called()
            mock_inference_poe.assert_called()


@patch("mm_poe.cli.set_seed")
@patch("mm_poe.cli.load_model")
@patch("mm_poe.cli.download_model")
@patch("mm_poe.cli.questionary.select")
@patch("mm_poe.cli.questionary.path")
@patch("mm_poe.cli.questionary.text")
def test_main_with_channel_below_average(
    mock_text,
    mock_path,
    mock_select,
    mock_download_model,
    mock_load_model,
    mock_set_seed,
):
    mock_select.return_value.ask.side_effect = [
        "BLIP2-OPT",  # args.model_family
        "Salesforce/blip2-opt-2.7b",  # args.checkpoint
        "FP16",  # args.loading_precision
        "channel",  # args.scoring_method_for_process_of_elimination
        "below_average",  # args.mask_strategy_for_process_of_elimination
        "1",  # args.label
    ]

    mock_path.return_value.ask.side_effect = [
        "./models/",  # args.output_dir
        "./images/image.png",  # args.image_path
    ]

    mock_text.return_value.ask.side_effect = [
        "Describe the image.",  # args.question
        "apple,banana,orange",  # args.choices
    ]

    mock_download_model.return_value = 0
    mock_model = MagicMock()
    mock_tokenizer = MagicMock()
    mock_load_model.return_value = (mock_model, mock_tokenizer)

    with patch("torch.device") as mock_device:
        mock_device.return_value = "cuda:0"

        with patch("mm_poe.cli.load_data") as mock_load_data, patch(
            "torch.utils.data.DataLoader"
        ) as mock_data_loader_class, patch(
            "mm_poe.cli.inference_language_modeling"
        ) as mock_inference_language_modeling, patch(
            "mm_poe.cli.inference_process_of_elimination"
        ) as mock_inference_poe, patch(
            "mm_poe.cli.compute_mask_process_of_elimination"
        ) as mock_compute_mask, patch(
            "mm_poe.cli.create_multiple_choice_prompt"
        ) as mock_create_mcp:

            mock_dataset = MagicMock()
            mock_dataset.map.return_value = mock_dataset
            mock_load_data.return_value = (
                ["hypothesis0", "hypothesis1", "hypothesis2"],
                "premise",
                "image_path",
                mock_dataset,
                mock_dataset,
            )

            mock_data_loader = MagicMock()
            mock_data_loader_class.return_value = mock_data_loader

            mock_inference_language_modeling.return_value = (
                torch.tensor([[0.3, 0.4, 0.3]]),
                None,
                None,
                torch.tensor([1]),
            )
            mock_inference_poe.return_value = (
                torch.tensor([[0.3, 0.4, 0.3]]),
                1.0,
                None,
                torch.tensor([1]),
            )

            mock_compute_mask.return_value = torch.tensor([[1, 0, 1]])

            def mock_create_mcp_fn(example, **kwargs):
                return example

            mock_create_mcp.side_effect = mock_create_mcp_fn

            main()

            mock_set_seed.assert_called_once_with(0)
            mock_download_model.assert_called()
            mock_load_model.assert_called()
            mock_load_data.assert_called()
            mock_inference_language_modeling.assert_called()
            mock_inference_poe.assert_called()


@patch("mm_poe.cli.set_seed")
@patch("mm_poe.cli.load_model")
@patch("mm_poe.cli.download_model")
@patch("mm_poe.cli.questionary.select")
@patch("mm_poe.cli.questionary.path")
@patch("mm_poe.cli.questionary.text")
def test_main_with_mask_strategy_min_k(
    mock_text,
    mock_path,
    mock_select,
    mock_download_model,
    mock_load_model,
    mock_set_seed,
):
    mock_select.return_value.ask.side_effect = [
        "GIT",
        "microsoft/git-base-vqav2",
        "FP32",
        "language_modeling",
        "min_k",
        "0",
    ]
    mock_path.return_value.ask.side_effect = [
        "./models/",
        "./images/image.png",
    ]
    mock_text.return_value.ask.side_effect = [
        "What is in the image?",
        "cat,dog,horse",
    ]
    mock_download_model.return_value = 0
    mock_model = MagicMock()
    mock_tokenizer = MagicMock()
    mock_load_model.return_value = (mock_model, mock_tokenizer)

    with patch("torch.device") as mock_device:
        mock_device.return_value = "cpu"

        # Modify args to include mask_token
        with patch("mm_poe.cli.Namespace") as mock_namespace:
            args = MagicMock()
            args.min_k = 1
            args.process_of_elimination_prompt = (
                "Select the most suitable option "
                + "to answer the question. Ignore [MASK] options."
            )
            mock_namespace.return_value = args

            with patch("mm_poe.cli.load_data") as mock_load_data, patch(
                "torch.utils.data.DataLoader"
            ) as mock_data_loader_class, patch(
                "mm_poe.cli.inference_language_modeling"
            ) as mock_inference_lm, patch(
                "mm_poe.cli.inference_process_of_elimination"
            ) as mock_inference_poe, patch(
                "mm_poe.cli.compute_mask_process_of_elimination"
            ) as mock_compute_mask, patch(
                "mm_poe.cli.create_multiple_choice_prompt"
            ) as mock_create_mcp:

                mock_dataset = MagicMock()
                mock_dataset.map.return_value = mock_dataset
                mock_load_data.return_value = (
                    ["hypothesis0", "hypothesis1", "hypothesis2"],
                    "premise",
                    "image_path",
                    mock_dataset,
                    mock_dataset,
                )

                mock_data_loader = MagicMock()
                mock_data_loader_class.return_value = mock_data_loader

                predictions = torch.tensor([[0.1, 0.2, 0.7]])
                masks = torch.tensor([[0, 1, 1]])
                mock_inference_lm.return_value = (
                    predictions,
                    None,
                    None,
                    torch.tensor([2]),
                )
                mock_inference_poe.return_value = (
                    predictions,
                    1.0,
                    None,
                    torch.tensor([2]),
                )
                mock_compute_mask.return_value = masks

                def mock_create_mcp_fn(example, **kwargs):
                    return example

                mock_create_mcp.side_effect = mock_create_mcp_fn

                main()
                mock_set_seed.assert_called_once_with(0)
                mock_load_model.assert_called()
                mock_load_data.assert_called()
                mock_compute_mask.assert_called_with(
                    predictions, "min_k", min_k=1
                )


@patch("mm_poe.cli.set_seed")
@patch("mm_poe.cli.load_model")
@patch("mm_poe.cli.download_model")
@patch("mm_poe.cli.questionary.select")
@patch("mm_poe.cli.questionary.path")
@patch("mm_poe.cli.questionary.text")
def test_main_with_mask_token(
    mock_text,
    mock_path,
    mock_select,
    mock_download_model,
    mock_load_model,
    mock_set_seed,
):
    mock_select.return_value.ask.side_effect = [
        "GIT",
        "microsoft/git-base-vqav2",
        "FP32",
        "language_modeling",
        "below_average",
        "0",
    ]
    mock_path.return_value.ask.side_effect = [
        "./models/",
        "./images/image.png",
    ]
    mock_text.return_value.ask.side_effect = [
        "What is in the image?",
        "cat,dog,horse",
    ]
    mock_download_model.return_value = 0
    mock_model = MagicMock()
    mock_tokenizer = MagicMock()
    mock_load_model.return_value = (mock_model, mock_tokenizer)

    with patch("torch.device") as mock_device:
        mock_device.return_value = "cpu"

        # Modify args to include mask_token
        with patch("mm_poe.cli.Namespace") as mock_namespace:
            args = MagicMock()
            args.mask_token = "XXX"
            args.process_of_elimination_prompt = (
                "Select the most suitable option "
                + "to answer the question. Ignore [MASK] options."
            )
            mock_namespace.return_value = args

            with patch("mm_poe.cli.load_data") as mock_load_data, patch(
                "torch.utils.data.DataLoader"
            ) as mock_data_loader_class, patch(
                "mm_poe.cli.inference_language_modeling"
            ) as mock_inference_lm, patch(
                "mm_poe.cli.inference_process_of_elimination"
            ) as mock_inference_poe, patch(
                "mm_poe.cli.compute_mask_process_of_elimination"
            ) as mock_compute_mask, patch(
                "mm_poe.cli.create_multiple_choice_prompt"
            ) as mock_create_mcp:

                mock_dataset = MagicMock()
                mock_dataset.map.return_value = mock_dataset
                mock_load_data.return_value = (
                    ["hypothesis0", "hypothesis1", "hypothesis2"],
                    "premise",
                    "image_path",
                    mock_dataset,
                    mock_dataset,
                )

                mock_data_loader = MagicMock()
                mock_data_loader_class.return_value = mock_data_loader

                predictions = torch.tensor([[0.1, 0.2, 0.7]])
                masks = torch.tensor([[0, 1, 1]])
                mock_inference_lm.return_value = (
                    predictions,
                    None,
                    None,
                    torch.tensor([2]),
                )
                mock_inference_poe.return_value = (
                    predictions,
                    1.0,
                    None,
                    torch.tensor([2]),
                )
                mock_compute_mask.return_value = masks

                def mock_create_mcp_fn(example, **kwargs):
                    assert "[MASK]" not in kwargs["multiple_choice_prompt"]
                    return example

                mock_create_mcp.side_effect = mock_create_mcp_fn

                main()
                mock_set_seed.assert_called_once_with(0)
                mock_load_model.assert_called()
                mock_load_data.assert_called()
                mock_compute_mask.assert_called_with(
                    predictions, "below_average"
                )


@patch("mm_poe.cli.set_seed")
@patch("mm_poe.cli.load_model")
@patch("mm_poe.cli.download_model")
@patch("mm_poe.cli.questionary.select")
@patch("mm_poe.cli.questionary.path")
@patch("mm_poe.cli.questionary.text")
def test_main_with_mask_strategy_min_k(
    mock_text,
    mock_path,
    mock_select,
    mock_download_model,
    mock_load_model,
    mock_set_seed,
):
    mock_select.return_value.ask.side_effect = [
        "GIT",
        "microsoft/git-base-vqav2",
        "FP32",
        "language_modeling",
        "min_k",
        "0",
    ]
    mock_path.return_value.ask.side_effect = [
        "./models/",
        "./images/image.png",
    ]
    mock_text.return_value.ask.side_effect = [
        "What is in the image?",
        "cat,dog,horse",
    ]
    mock_download_model.return_value = 0
    mock_model = MagicMock()
    mock_tokenizer = MagicMock()
    mock_load_model.return_value = (mock_model, mock_tokenizer)

    with patch("torch.device") as mock_device:
        mock_device.return_value = "cpu"

        # Modify args to include mask_token
        with patch("mm_poe.cli.Namespace") as mock_namespace:
            args = MagicMock()
            args.min_k = 10
            args.process_of_elimination_prompt = (
                "Select the most suitable "
                + "option to answer the question. Ignore [MASK] options."
            )
            mock_namespace.return_value = args

            with patch("mm_poe.cli.load_data") as mock_load_data, patch(
                "torch.utils.data.DataLoader"
            ) as mock_data_loader_class, patch(
                "mm_poe.cli.inference_language_modeling"
            ) as mock_inference_lm, patch(
                "mm_poe.cli.inference_process_of_elimination"
            ) as mock_inference_poe, patch(
                "mm_poe.cli.compute_mask_process_of_elimination"
            ) as mock_compute_mask, patch(
                "mm_poe.cli.create_multiple_choice_prompt"
            ) as mock_create_mcp:

                mock_dataset = MagicMock()
                mock_dataset.map.return_value = mock_dataset
                mock_load_data.return_value = (
                    ["hypothesis0", "hypothesis1", "hypothesis2"],
                    "premise",
                    "image_path",
                    mock_dataset,
                    mock_dataset,
                )

                mock_data_loader = MagicMock()
                mock_data_loader_class.return_value = mock_data_loader

                predictions = torch.tensor([[0.1, 0.2, 0.7]])
                masks = torch.tensor([[0, 1, 1]])
                mock_inference_lm.return_value = (
                    predictions,
                    None,
                    None,
                    torch.tensor([2]),
                )
                mock_inference_poe.return_value = (
                    predictions,
                    1.0,
                    None,
                    torch.tensor([2]),
                )
                mock_compute_mask.return_value = masks

                def mock_create_mcp_fn(example, **kwargs):
                    return example

                mock_create_mcp.side_effect = mock_create_mcp_fn

                main()
                mock_set_seed.assert_called_once_with(0)
                mock_load_model.assert_called()
                mock_load_data.assert_called()
                mock_compute_mask.assert_called_with(
                    predictions, "min_k", min_k=2
                )



from tqdm import tqdm

import torch
import torch.nn.functional as F
from torch.utils.data import DataLoader
from transformers import GenerationConfig


def inference_language_modeling_old(model, eval_dataloader, device):
    model.eval()
    predictions = torch.zeros(0)
    labels = torch.zeros(0)
    torch.cuda.empty_cache()

    pbar = tqdm(eval_dataloader, desc="Inference")
    for batch in pbar:
        # e.g., (batch_size, #option, ending_seq_len): (32, 2, 18)
        ending_shape = batch["ending_input_ids"].shape
        # flatten
        header_input_ids = (
            batch["header_input_ids"]
            .view(-1, batch["header_input_ids"].shape[-1])
            .to(device)
        )
        ending_input_ids = (
            batch["ending_input_ids"]
            .view(-1, batch["ending_input_ids"].shape[-1])
            .to(device)
        )

        # adding this line of code takes me more than an hour.
        # without adding torch.no_grad, GPU usage will muiltply by 4.
        with torch.no_grad():
            outputs = model(
                input_ids=header_input_ids, labels=ending_input_ids
            )

        _, logits = outputs.loss, outputs.logits
        # e.g., (batch_size * #option, ending_seq_len, #vocab): (64, 18, 32128)
        logits = logits.view(-1, logits.shape[-1])
        # ignore padding token: 0
        ce_loss = (
            F.cross_entropy(
                logits,
                ending_input_ids.view(-1),
                reduction="none",
                ignore_index=0,
            )
            .detach()
            .cpu()
        )
        # each score is the negative log-likelihood of a ending given a header.
        batch_predictions = (
            ce_loss.view(ending_shape).sum(dim=-1).argmin(dim=-1)
        )
        batch_labels = batch["label"]
        predictions = torch.cat((predictions, batch_predictions))
        labels = torch.cat((labels, batch_labels))

        # make accuracy accumulative
        batch_accuracy = (
            batch_predictions == batch_labels
        ).sum().item() / len(batch_labels)
        total_accuracy = (predictions == labels).sum().item() / len(labels)
        pbar.set_description(
            f"Total Accuracy: {total_accuracy:.4f}, "
            + f"Batch Accuracy: {batch_accuracy:.4f}"
        )
    return total_accuracy


def inference_contrastive_decoding_old(
    amateur_model, expert_model, eval_dataloader, device
):
    amateur_model.eval()
    expert_model.eval()
    predictions = torch.zeros(0)
    labels = torch.zeros(0)
    torch.cuda.empty_cache()

    pbar = tqdm(eval_dataloader, desc="Inference")
    for batch in pbar:
        # e.g., (batch_size, #option, ending_seq_len): (32, 2, 18)
        ending_shape = batch["ending_input_ids"].shape
        # flatten
        header_input_ids = (
            batch["header_input_ids"]
            .view(-1, batch["header_input_ids"].shape[-1])
            .to(device)
        )
        ending_input_ids = (
            batch["ending_input_ids"]
            .view(-1, batch["ending_input_ids"].shape[-1])
            .to(device)
        )

        # key step: compute logits.
        with torch.no_grad():
            amateur_model_logits = amateur_model(
                input_ids=header_input_ids, labels=ending_input_ids
            ).logits
            expert_model_logits = expert_model(
                input_ids=header_input_ids, labels=ending_input_ids
            ).logits

        logits = expert_model_logits - amateur_model_logits
        # e.g., (batch_size * #option, ending_seq_len, #vocab): (64, 18, 32128)
        logits = logits.view(-1, logits.shape[-1])
        # ignore padding token: 0
        ce_loss = (
            F.cross_entropy(
                logits,
                ending_input_ids.view(-1),
                reduction="none",
                ignore_index=0,
            )
            .detach()
            .cpu()
        )
        # each score is the negative log-likelihood of a ending given a header.
        batch_predictions = (
            ce_loss.view(ending_shape).sum(dim=-1).argmin(dim=-1)
        )
        batch_labels = batch["label"]
        predictions = torch.cat((predictions, batch_predictions))
        labels = torch.cat((labels, batch_labels))

        # make accuracy accumulative
        batch_accuracy = (
            batch_predictions == batch_labels
        ).sum().item() / len(batch_labels)
        total_accuracy = (predictions == labels).sum().item() / len(labels)
        pbar.set_description(
            f"Total Accuracy: {total_accuracy:.4f}, "
            + f"Batch Accuracy: {batch_accuracy:.4f}"
        )
    return total_accuracy


def inference_language_modeling(
    model, eval_dataloader, device, compute_func, pad_token_id
):
    model.eval()
    lm_predictions = torch.zeros(0)
    avg_lm_predictions = torch.zeros(0)
    labels = torch.zeros(0)
    torch.cuda.empty_cache()
    avg_log_probs = []

    pbar = tqdm(eval_dataloader, desc="Inference")
    for batch in pbar:
        log_prob = compute_func(batch, model, device, pad_token_id)
        avg_log_prob = log_prob / batch["ending_attention_mask"].sum(dim=-1)
        avg_log_probs.append(avg_log_prob)

        batch_predictions = log_prob.argmin(dim=-1)
        batch_avg_predictions = avg_log_prob.argmin(dim=-1)

        batch_labels = batch["label"]
        lm_predictions = torch.cat((lm_predictions, batch_predictions))
        avg_lm_predictions = torch.cat(
            (avg_lm_predictions, batch_avg_predictions)
        )
        labels = torch.cat((labels, batch_labels))

        # make accuracy accumulative
        lm_accuracy = (lm_predictions == labels).sum().item() / len(labels)
        avg_lm_accuracy = (avg_lm_predictions == labels).sum().item() / len(
            labels
        )
        pbar.set_description(
            f"Language modeling accuracy: {lm_accuracy:.4f}, "
            + f"Average language modeling accuracy: {avg_lm_accuracy:.4f}"
        )
    avg_log_probs = torch.cat(avg_log_probs, dim=0)
    return avg_log_probs, lm_accuracy, avg_lm_accuracy, lm_predictions


def inference_generate_synonyms(
    model,
    eval_dataloader,
    device,
    compute_func,
    pad_token_id,
    num_of_options,
    num_of_synonyms,
):
    model.eval()
    lm_predictions = torch.zeros(0)
    avg_lm_predictions = torch.zeros(0)
    labels = torch.zeros(0)
    torch.cuda.empty_cache()
    avg_log_probs = []

    pbar = tqdm(eval_dataloader, desc="Inference")
    for batch in pbar:
        log_prob = compute_func(batch, model, device, pad_token_id)
        avg_log_prob = log_prob / batch["ending_attention_mask"].sum(dim=-1)
        avg_log_probs.append(avg_log_prob)

        # need to aggregate according to original options.
        # each row in log_prob correspond to: h0, h1, ..., hn,
        # h0s0, h0s1, ..., h1s0, ...,

        # indexing log_prob to rearrange rows by keeping
        # options with corresponding synonyms together.
        log_prob = aggregate_optionw_with_synonyms(
            log_prob, num_of_options, num_of_synonyms
        )
        avg_log_prob = aggregate_optionw_with_synonyms(
            avg_log_prob, num_of_options, num_of_synonyms
        )
        # then reshape, and then aggregate options and synonyms by averaging.
        log_prob = log_prob.view(-1, num_of_options, num_of_synonyms + 1).mean(
            dim=-1
        )
        avg_log_prob = avg_log_prob.view(
            -1, num_of_options, num_of_synonyms + 1
        ).mean(dim=-1)

        batch_predictions = log_prob.argmin(dim=-1)
        batch_avg_predictions = avg_log_prob.argmin(dim=-1)

        batch_labels = batch["label"]
        lm_predictions = torch.cat((lm_predictions, batch_predictions))
        avg_lm_predictions = torch.cat(
            (avg_lm_predictions, batch_avg_predictions)
        )
        labels = torch.cat((labels, batch_labels))

        # make accuracy accumulative
        lm_accuracy = (lm_predictions == labels).sum().item() / len(labels)
        avg_lm_accuracy = (avg_lm_predictions == labels).sum().item() / len(
            labels
        )
        pbar.set_description(
            f"Language modeling accuracy: {lm_accuracy:.4f}, "
            + f"Average language modeling accuracy: {avg_lm_accuracy:.4f}"
        )
    avg_log_probs = torch.cat(avg_log_probs, dim=0)
    return avg_log_probs, lm_accuracy, avg_lm_accuracy, lm_predictions


def inference_calibration(
    model,
    eval_dataloader,
    eval_calibration_dataloader,
    device,
    compute_func,
    pad_token_id,
):
    model.eval()
    lm_predictions = torch.zeros(0)
    avg_lm_predictions = torch.zeros(0)
    labels = torch.zeros(0)
    torch.cuda.empty_cache()
    avg_log_probs = []

    pbar = tqdm(
        zip(eval_dataloader, eval_calibration_dataloader),
        desc="Inference",
        total=len(eval_dataloader),
    )
    for batch, batch_calibration in pbar:
        log_prob = compute_func(batch, model, device, pad_token_id)
        log_prob_calibration = compute_func(
            batch_calibration, model, device, pad_token_id
        )
        log_prob = log_prob - log_prob_calibration
        avg_log_prob = log_prob / batch["ending_attention_mask"].sum(dim=-1)
        avg_log_probs.append(avg_log_prob)

        batch_predictions = log_prob.argmin(dim=-1)
        batch_avg_predictions = avg_log_prob.argmin(dim=-1)

        batch_labels = batch["label"]
        lm_predictions = torch.cat((lm_predictions, batch_predictions))
        avg_lm_predictions = torch.cat(
            (avg_lm_predictions, batch_avg_predictions)
        )
        labels = torch.cat((labels, batch_labels))

        # make accuracy accumulative
        lm_accuracy = (lm_predictions == labels).sum().item() / len(labels)
        avg_lm_accuracy = (avg_lm_predictions == labels).sum().item() / len(
            labels
        )
        pbar.set_description(
            f"Calibration accuracy: {lm_accuracy:.4f}, "
            + f"Average calibration accuracy: {avg_lm_accuracy:.4f}"
        )
    avg_log_probs = torch.cat(avg_log_probs, dim=0)
    return avg_log_probs, lm_accuracy, avg_lm_accuracy, lm_predictions


def inference_contrastive_decoding(method, model, **kwargs):
    args = kwargs["args"]
    raw_dataset = kwargs["raw_dataset"]
    device = kwargs["device"]
    compute_func = kwargs["compute_func"]
    tokenizer = kwargs["tokenizer"]
    processor = kwargs["processor"]
    ending_names = kwargs["ending_names"]
    header_name = kwargs["header_name"]
    image_header_name = kwargs["image_header_name"]
    preprocess_func = kwargs["preprocess_func"]
    preprocess_func_channel = kwargs["preprocess_func_channel"]

    fn_kwargs = {
        "ending_names": ending_names,
        "header_name": header_name,
        "tokenizer": tokenizer,
    }
    if args.model_family in [
        "BLIP2-OPT",
        "BLIP2-T5",
        "InstructBLIP",
        "GIT",
        "PaliGemma",
        "Idefics2",
    ]:
        fn_kwargs = {
            "ending_names": ending_names,
            "header_name": header_name,
            "tokenizer": tokenizer,
            "processor": processor,
            "image_header_name": image_header_name,
        }
    # num_of_options = len(ending_names)
    tokenized_dataset = raw_dataset.map(
        preprocess_func,
        fn_kwargs=fn_kwargs,
        batched=True,
        batch_size=args.batch_size,
    )
    eval_dataloader = DataLoader(
        tokenized_dataset, batch_size=args.batch_size, shuffle=False
    )
    if method in ["language_modeling", "multiple_choice_prompt"]:
        avg_log_probs, lm_accuracy, avg_lm_accuracy, lm_predictions = (
            inference_language_modeling(
                model,
                eval_dataloader,
                device,
                compute_func,
                tokenizer.pad_token_id,
            )
        )
    elif method == "calibration":
        fn_kwargs = {
            "ending_names": ending_names,
            "header_name": "uncond_premise",  # the difference is here
            "tokenizer": tokenizer,
        }
        if args.model_family in [
            "BLIP2-OPT",
            "BLIP2-T5",
            "InstructBLIP",
            "GIT",
            "PaliGemma",
            "Idefics2",
        ]:
            fn_kwargs = {
                "ending_names": ending_names,
                "header_name": "uncond_premise",
                "tokenizer": tokenizer,
                "processor": processor,
                "image_header_name": image_header_name,
            }
        tokenized_calibration_dataset = raw_dataset.map(
            preprocess_func,
            fn_kwargs=fn_kwargs,
            batched=True,
            batch_size=args.batch_size,
        )
        eval_calibration_dataloader = DataLoader(
            tokenized_calibration_dataset,
            batch_size=args.batch_size,
            shuffle=False,
        )
        avg_log_probs, lm_accuracy, avg_lm_accuracy, lm_predictions = (
            inference_calibration(
                model,
                eval_dataloader,
                eval_calibration_dataloader,
                device,
                compute_func,
                tokenizer.pad_token_id,
            )
        )
    elif method == "channel":
        # simple solution: swap first sentence and
        # second sentence in both preprocessing functions
        tokenized_channel_dataset = raw_dataset.map(
            preprocess_func_channel,
            fn_kwargs=fn_kwargs,
            batched=True,
            batch_size=args.batch_size,
        )
        eval_channel_dataloader = DataLoader(
            tokenized_channel_dataset,
            batch_size=args.batch_size,
            shuffle=False,
        )
        avg_log_probs, lm_accuracy, avg_lm_accuracy, lm_predictions = (
            inference_language_modeling(
                model,
                eval_channel_dataloader,
                device,
                compute_func,
                tokenizer.pad_token_id,
            )
        )
    else:
        raise NotImplementedError
    return avg_log_probs, lm_accuracy, avg_lm_accuracy, lm_predictions


def compute_mask_process_of_elimination(
    avg_log_probs, mask_strategy, **kwargs
):
    masks = torch.ones_like(avg_log_probs)
    if mask_strategy == "lowest":
        # soft masking (v1), i.e., get rid of the least likely answer.
        masks[
            torch.arange(avg_log_probs.shape[0]), avg_log_probs.argmax(dim=-1)
        ] = 0
    elif mask_strategy == "below_average":
        # v2: Calculate the row-wise mean
        row_mean = avg_log_probs.mean(dim=1, keepdim=True)
        # Set values below the mean to 0
        masks[avg_log_probs > row_mean] = 0
    elif mask_strategy == "lowest_iter":
        # similar to lowest, but ignore inf,
        # and mask from the remaining options.
        # soft masking (v1), i.e., get rid of the least likely answer.
        avg_log_probs[avg_log_probs == float("inf")] = float("-inf")
        masks[
            torch.arange(avg_log_probs.shape[0]), avg_log_probs.argmax(dim=-1)
        ] = 0
        # set mask that correspond to inf to 0
        masks[avg_log_probs == float("-inf")] = 0
    elif mask_strategy == "min_k":
        min_k = kwargs["min_k"]
        # keep the min k options
        avg_log_probs_f32 = avg_log_probs.float()
        _, min_k_indices = avg_log_probs_f32.topk(min_k, dim=-1)
        masks[
            torch.arange(avg_log_probs_f32.shape[0]).unsqueeze(-1),
            min_k_indices,
        ] = 0
    else:
        raise NotImplementedError
    return masks


def inference_process_of_elimination(
    model, eval_dataloader, device, compute_func, pad_token_id
):
    model.eval()
    lm_predictions = torch.zeros(0)
    avg_lm_predictions = torch.zeros(0)
    labels = torch.zeros(0)
    torch.cuda.empty_cache()
    avg_log_probs = []

    pbar = tqdm(eval_dataloader, desc="Inference")
    for batch in pbar:
        log_prob = compute_func(batch, model, device, pad_token_id)
        # apply hard masking
        log_prob[batch["mask"] == 0] = float("inf")
        avg_log_prob = log_prob / batch["ending_attention_mask"].sum(dim=-1)
        avg_log_probs.append(avg_log_prob)

        batch_predictions = log_prob.argmin(dim=-1)
        batch_avg_predictions = avg_log_prob.argmin(dim=-1)

        batch_labels = batch["label"]
        lm_predictions = torch.cat((lm_predictions, batch_predictions))
        avg_lm_predictions = torch.cat(
            (avg_lm_predictions, batch_avg_predictions)
        )
        labels = torch.cat((labels, batch_labels))

        # make accuracy accumulative
        lm_accuracy = (lm_predictions == labels).sum().item() / len(labels)
        avg_lm_accuracy = (avg_lm_predictions == labels).sum().item() / len(
            labels
        )
        pbar.set_description(
            f"Process of elimination accuracy: {lm_accuracy:.4f}, "
            + f"Average process of elimination accuracy: {avg_lm_accuracy:.4f}"
        )
    avg_log_probs = torch.cat(avg_log_probs, dim=0)
    return avg_log_probs, lm_accuracy, avg_lm_accuracy, lm_predictions


def compute_conditional_score_seq2seq(batch, model, device, pad_token_id):
    # returns log_prob of p(y|x) for each batch

    # e.g., (batch_size, #option, ending_seq_len): (32, 2, 18)
    ending_shape = batch["ending_input_ids"].shape
    # flatten. both input_ids has 0 as padding token.
    header_input_ids = (
        batch["header_input_ids"]
        .view(-1, batch["header_input_ids"].shape[-1])
        .to(device)
    )
    header_attention_mask = (
        batch["header_attention_mask"]
        .view(-1, batch["header_attention_mask"].shape[-1])
        .to(device)
    )
    ending_input_ids = (
        batch["ending_input_ids"]
        .view(-1, batch["ending_input_ids"].shape[-1])
        .to(device)
    )

    # adding this line of code takes me more than an hour.
    # without adding torch.no_grad, GPU usage will muiltply by 4.
    with torch.no_grad():
        outputs = model(
            input_ids=header_input_ids,
            attention_mask=header_attention_mask,
            labels=ending_input_ids,
        )

    _, logits = outputs.loss, outputs.logits
    # e.g., (batch_size * #option, ending_seq_len, #vocab): (64, 18, 32128)
    logits = logits.view(-1, logits.shape[-1])
    # ignore padding token: 0
    ce_loss = (
        F.cross_entropy(
            logits,
            ending_input_ids.view(-1),
            reduction="none",
            ignore_index=pad_token_id,
        )
        .detach()
        .cpu()
    )
    # each score is the negative log-likelihood of a ending given a header.
    # batch_predictions = ce_loss.view(ending_shape).sum(dim=-1).argmin(dim=-1)
    log_prob = ce_loss.view(ending_shape).sum(dim=-1)
    return log_prob


def compute_conditional_score_causal(batch, model, device, pad_token_id):
    # returns log_prob of p(y|x) for each batch
    # make sure the padding token is aligned with tokenizer.pad_token_id
    # and preprocess_function_causal
    # padding_token = 50256

    input_ids = (
        batch["input_ids"].view(-1, batch["input_ids"].shape[-1]).to(device)
    )
    labels = batch["labels"].view(-1, batch["labels"].shape[-1]).to(device)

    # adding this line of code takes me more than an hour.
    # without adding torch.no_grad, GPU usage will muiltply by 4.
    with torch.no_grad():
        outputs = model(
            input_ids=input_ids,
            # attention_mask = attention_mask,
            labels=labels,
        )

    _, logits = outputs.loss, outputs.logits
    # shift
    logits = logits[:, :-1].contiguous()
    labels = labels[:, 1:].contiguous()
    # e.g., (batch_size * #option, ending_seq_len, #vocab): (64, 18, 32128)
    logits = logits.view(-1, logits.shape[-1])
    # ignore padding token: 50256
    ce_loss = (
        F.cross_entropy(
            logits,
            labels.view(-1),
            reduction="none",
            ignore_index=pad_token_id,
        )
        .detach()
        .cpu()
    )
    # each score is the negative log-likelihood of a ending given a header.
    log_prob = ce_loss.view(
        batch["input_ids"].shape[0], batch["input_ids"].shape[1], -1
    ).sum(dim=-1)
    return log_prob


def compute_conditional_score_seq2seq_vqa(batch, model, device, pad_token_id):
    # returns log_prob of p(y|x) for each batch

    # e.g., (batch_size, #option, ending_seq_len): (32, 2, 18)
    ending_shape = batch["ending_input_ids"].shape
    # flatten. both input_ids has 0 as padding token.
    header_input_ids = (
        batch["header_input_ids"]
        .view(-1, batch["header_input_ids"].shape[-1])
        .to(device)
    )
    header_attention_mask = (
        batch["header_attention_mask"]
        .view(-1, batch["header_attention_mask"].shape[-1])
        .to(device)
    )
    ending_input_ids = (
        batch["ending_input_ids"]
        .view(-1, batch["ending_input_ids"].shape[-1])
        .to(device)
    )
    images = (
        batch["images"]
        .view(
            -1,
            batch["images"].shape[-3],
            batch["images"].shape[-2],
            batch["images"].shape[-1],
        )
        .to(device)
    )

    # adding this line of code takes me more than an hour.
    # without adding torch.no_grad, GPU usage will muiltply by 4.
    with torch.no_grad():
        outputs = model(
            input_ids=header_input_ids,
            attention_mask=header_attention_mask,
            pixel_values=images,
            labels=ending_input_ids,
        )

    _, logits = outputs.loss, outputs.logits
    logits = logits[:, : ending_input_ids.shape[-1], :]
    # e.g., (batch_size * #option, ending_seq_len, #vocab): (64, 18, 32128)
    logits = logits.contiguous().view(-1, logits.shape[-1])
    # ignore padding token: 0
    ce_loss = (
        F.cross_entropy(
            logits,
            ending_input_ids.view(-1),
            reduction="none",
            ignore_index=pad_token_id,
        )
        .detach()
        .cpu()
    )
    # each score is the negative log-likelihood of a ending given a header.
    # batch_predictions = ce_loss.view(ending_shape).sum(dim=-1).argmin(dim=-1)
    log_prob = ce_loss.view(ending_shape).sum(dim=-1)
    return log_prob


def compute_conditional_score_causal_vqa(batch, model, device, pad_token_id):
    # returns log_prob of p(y|x) for each batch
    # make sure the padding token is aligned with tokenizer.pad_token_id
    # and preprocess_function_causal
    # padding_token = 50256
    input_ids = (
        batch["input_ids"].view(-1, batch["input_ids"].shape[-1]).to(device)
    )
    header_attention_mask = (
        batch["header_attention_mask"]
        .view(-1, batch["header_attention_mask"].shape[-1])
        .to(device)
    )
    # ending_attention_mask = (
    #     batch["ending_attention_mask"]
    #     .view(-1, batch["ending_attention_mask"].shape[-1])
    #     .to(device)
    # )
    labels = batch["labels"].view(-1, batch["labels"].shape[-1]).to(device)
    images = (
        batch["images"]
        .view(
            -1,
            batch["images"].shape[-3],
            batch["images"].shape[-2],
            batch["images"].shape[-1],
        )
        .to(device)
    )

    # adding this line of code takes me more than an hour.
    # without adding torch.no_grad, GPU usage will muiltply by 4.
    with torch.no_grad():
        outputs = model(
            input_ids=input_ids,
            pixel_values=images,
            attention_mask=header_attention_mask,
            labels=labels,
        )

    _, logits = outputs.loss, outputs.logits
    logits = logits[:, -labels.shape[-1] :, :]  # for GIT

    # shift
    logits = logits[:, :-1].contiguous()
    labels = labels[:, 1:].contiguous()

    # e.g., (batch_size * #option, ending_seq_len, #vocab): (64, 18, 32128)
    logits = logits.view(-1, logits.shape[-1])
    # ignore padding token: 50256
    ce_loss = (
        F.cross_entropy(
            logits,
            labels.view(-1),
            reduction="none",
            ignore_index=pad_token_id,
        )
        .detach()
        .cpu()
    )
    # each score is the negative log-likelihood of a ending given a header.
    log_prob = ce_loss.view(
        batch["input_ids"].shape[0], batch["input_ids"].shape[1], -1
    ).sum(dim=-1)
    return log_prob


def generate_synonyms(args, model, tokenizer, tokenized_dataset):

    generation_config = GenerationConfig(
        max_new_tokens=50,
        do_sample=True,
        temperature=0.7,
        num_return_sequences=args.number_of_synonyms,
    )

    # get all columns of tokenized_dataset that starts with "hypothesis"
    hypothesis_columns = [
        col
        for col in tokenized_dataset.column_names
        if col.startswith("hypothesis")
    ]
    # batch inference? May check SEQA code or HF doc.
    synonyms_dict = {}
    for col in tqdm(hypothesis_columns, desc="Generate synonyms"):
        for option in tqdm(
            tokenized_dataset[col], desc=f"Generate synonyms for {col}"
        ):
            # prompt = f"Generate a synonym to '{option}':"
            prompt = args.generate_synonyms_prompt.replace(
                "'{option}'", option
            )
            inputs = tokenizer(prompt, return_tensors="pt").to(model.device)
            outputs = model.generate(
                **inputs, generation_config=generation_config
            )
            synonyms = tokenizer.batch_decode(
                outputs, skip_special_tokens=True
            )
            # store the synonyms, so map() is easy.
            # directly mapping here? All I need to do
            # is to create duplicates of the instance with synonyms.
            synonyms_dict[option] = synonyms

    return synonyms_dict


def aggregate_optionw_with_synonyms(tensor, num_of_options, num_of_synonyms):
    # this function changes the column order.
    # tensor: (batch_size, num_of_options * (num_of_synonyms + 1))
    old_index = list(range(tensor.shape[1]))
    aggregated_index = [-1] * len(old_index)
    # exapmle: commonsenseqa 5 options with 3 synonyms: 0, 4, 8, 12, 16
    options_index_old = list(range(num_of_options))  # e.g., 0..4
    options_index_new = [
        i * (num_of_synonyms + 1) for i in options_index_old
    ]  # e.g., 0, 4, 8, 12, 16
    remain_index = [
        i for i in old_index if i not in options_index_old
    ]  # e.g., 5..19
    for i, _ in enumerate(aggregated_index):
        if i in options_index_new:  # 0, 4, 8, 12, 16
            aggregated_index[i] = options_index_old.pop(0)
        else:
            aggregated_index[i] = remain_index.pop(0)

    # aggregated_index = options_index + \
    #     [i for i in old_index if i not in options_index]
    tensor[:, old_index] = tensor[:, aggregated_index]
    return tensor



# test_data.py

import os
import sys
import torch
import pytest
from unittest import mock
from unittest.mock import MagicMock, patch, Mock
from torch.utils.data import DataLoader
from tqdm import tqdm
import json
import xml.etree.ElementTree as ET
from PIL import Image
import random

# Import all functions from data.py
# Adjust the import path based on your project structure
# Assuming data.py is in the same directory as test_data.py
from mm_poe.methods.utils.data import (
    upload_to_huggingface_hub,
    preprocess_function_seq2seq,
    preprocess_function_causal,
    preprocess_function_seq2seq_vqa,
    preprocess_function_causal_vqa,
    preprocess_function_seq2seq_channel,
    preprocess_function_causal_channel,
    preprocess_function_seq2seq_vqa_channel,
    preprocess_function_causal_vqa_channel,
    create_multiple_choice_prompt,
    create_synonym_dataset,
    copa_loader,
    cqa_loader,
    obqa_loader,
    piqa_loader,
    qasc_loader,
    siqa_loader,
    winogrande_loader,
    date_understanding_loader,
    anli_loader,
    generate_n_shot_demonstrations,
    create_n_shot_splits,
    generate_n_shot_poe_demonstrations,
    vqa_loader,
    scienceqa_loader,
    ai2d_loader,
    single_inference_loader,
)


# Mock class for argparse.Namespace
class Args:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


@pytest.fixture
def sample_args():
    return Args(
        dataset="test_dataset",
        seed=42,
        n_shot=5,
        sample=10,
        checkpoint="checkpoints/test_checkpoint",
        batch_size=32,
        method="test_method",
        ending_names=["choice0", "choice1", "choice2", "choice3"],
        header_name="question",
        tokenizer=Mock(),
        processor=Mock(),
        image_header_name="image_path",
        multiple_choice_prompt="Please select the correct answer:",
        scoring_method="other_method",
        num_of_options=3,
        mask_token=None,
        number_of_synonyms=2,
        calibration_prompt=" the answer is:",
        num_options=4,
        mask=[1, 0, 1],
        # sample=None,
        n_shot_demonstrations="",
        image_processor=Mock(),
        synonyms_dict={
            "Paris": ["Paris1", "Paris2"],
            "London": ["London1", "London2"],
        },
        question="What is the capital of France?",
        choices=["Paris", "London", "Berlin"],
        label=0,
    )


def test_upload_to_huggingface_hub(sample_args):
    dataset = MagicMock()
    args = sample_args
    suffix = f"{args.dataset}_{args.seed}_{args.n_shot}_{args.sample}_{args.checkpoint.split('/')[-1]}_{args.batch_size}"
    temp_data_path = os.path.join(f"../temp_data/{args.method}", suffix)

    with patch("os.system") as mock_system:
        upload_to_huggingface_hub(dataset, args)
        dataset.save_to_disk.assert_called_once_with(temp_data_path)


def test_preprocess_function_seq2seq(sample_args):
    examples = {
        "question": ["What is the capital of France?", "What is 2+2?"],
        "choice0": ["Paris", "3"],
        "choice1": ["London", "4"],
        "choice2": ["Berlin", "5"],
        "choice3": ["Madrid", "6"],
    }
    tokenizer = MagicMock()
    # Adjust the tokenizer mock to return the correct number of tokens
    tokenizer.return_value = {
        "input_ids": [[i] for i in range(8)],  # 2 questions * 4 choices = 8
        "attention_mask": [[1] for _ in range(8)],
    }
    kwargs = {
        "ending_names": ["choice0", "choice1", "choice2", "choice3"],
        "header_name": "question",
        "tokenizer": tokenizer,
    }
    output = preprocess_function_seq2seq(examples, **kwargs)
    assert "header_input_ids" in output
    assert "header_attention_mask" in output
    assert "ending_input_ids" in output
    assert "ending_attention_mask" in output
    num_choice = 4
    for key in output:
        assert len(output[key]) == len(examples["question"])
        for sublist in output[key]:
            assert len(sublist) == num_choice


def test_preprocess_function_causal(sample_args):
    examples = {
        "question": ["What is the capital of France?"],
        "choice0": ["Paris"],
        "choice1": ["London"],
    }
    tokenizer = MagicMock()
    tokenizer.pad_token_id = 0
    # Fix the lambda function to define 's'
    tokenizer.side_effect = lambda x, truncation: (
        {
            "input_ids": [list(range(len(s))) for s in x],
            "attention_mask": [[1] * len(s) for s in x],
        }
        if isinstance(x, list)
        else {}
    )
    kwargs = {
        "ending_names": ["choice0", "choice1"],
        "header_name": "question",
        "tokenizer": tokenizer,
    }
    output = preprocess_function_causal(examples, **kwargs)
    assert "input_ids" in output
    assert "labels" in output
    assert "ending_attention_mask" in output


def test_preprocess_function_seq2seq_vqa(sample_args):
    examples = {
        "question": ["What is shown in the image?"],
        "choice0": ["Cat"],
        "choice1": ["Dog"],
        "image_path": ["path/to/image1.jpg"],
    }
    processor = MagicMock()
    # Adjust the tokenizer and image processor mocks
    processor.tokenizer.return_value = {
        "input_ids": [[i] for i in range(2)],
        "attention_mask": [[1] for _ in range(2)],
    }
    data_obj = MagicMock()
    data_obj.data = {
        "pixel_values": torch.tensor(
            [[[1, 2], [3, 4]]] * 2
        )  # Repeat to match the number of choices
    }
    processor.image_processor.return_value = data_obj
    kwargs = {
        "ending_names": ["choice0", "choice1"],
        "header_name": "question",
        "image_header_name": "image_path",
        "processor": processor,
        "image_token": "",
    }
    with patch("PIL.Image.open", return_value=MagicMock(spec=Image.Image)):
        output = preprocess_function_seq2seq_vqa(examples, **kwargs)
        assert "header_input_ids" in output
        assert "ending_input_ids" in output
        assert "images" in output
        assert len(output["images"]) == len(examples["question"])
        for img_list in output["images"]:
            assert len(img_list) == len(kwargs["ending_names"])


def test_create_multiple_choice_prompt(sample_args):
    example = {
        "premise": "What is the capital of France?",
        "uncond_premise": "The answer is:",
        "hypothesis0": "Paris",
        "hypothesis1": "London",
        "hypothesis2": "Berlin",
        "mask": [1, 0, 1],
    }
    kwargs = {
        "multiple_choice_prompt": "Please choose the correct answer:",
        "scoring_method": "other_method",
        "num_of_options": 3,
        "mask_token": None,
    }
    output = create_multiple_choice_prompt(example, **kwargs)
    expected_premise = "Please choose the correct answer:\n Question: What is the capital of France?\nA. Paris\nB. [MASK]\nC. Berlin\nAnswer:"
    assert output["premise"] == expected_premise
    kwargs["scoring_method"] = "multiple_choice_prompt"
    example["premise"] = (
        " Question: What is the capital of France?\nA. Paris\nB. London\nC. Berlin\nAnswer:"
    )
    output = create_multiple_choice_prompt(example, **kwargs)
    assert output["premise"] == expected_premise


def test_create_synonym_dataset(sample_args):
    examples = {
        "hypothesis0": ["Paris", "London"],
        "hypothesis1": ["Berlin", "Madrid"],
    }
    kwargs = {
        "args": sample_args,
        "synonyms_dict": {
            "Paris": ["Paris1", "Paris2"],
            "London": ["London1", "London2"],
            "Berlin": ["Berlin1", "Berlin2"],
            "Madrid": ["Madrid1", "Madrid2"],
        },
    }
    output = create_synonym_dataset(examples, **kwargs)
    for hypothesis in ["hypothesis0", "hypothesis1"]:
        for i in range(sample_args.number_of_synonyms):
            key = f"{hypothesis}_synonyms_{i}"
            assert key in output
            assert len(output[key]) == len(examples[hypothesis])


def test_copa_loader(sample_args):
    args = sample_args
    args.multiple_choice_prompt = None
    xml_content = """<root>
    <item most-plausible-alternative="1" asks-for="effect">
        <p>It started to rain.</p>
        <a1>I opened my umbrella.</a1>
        <a2>I wore sunglasses.</a2>
    </item>
    </root>"""
    with patch("xml.etree.ElementTree.parse") as mock_parse:
        mock_tree = ET.ElementTree(ET.fromstring(xml_content))
        mock_parse.return_value = mock_tree
        examples = copa_loader("dummy_path.xml", args)
        assert len(examples) == 1
        assert examples[0]["label"] == 0
        assert examples[0]["premise"] == " It started to rain so"
        assert examples[0]["hypothesis0"] == " i opened my umbrella."
        assert examples[0]["hypothesis1"] == " i wore sunglasses."


def test_copa_loader_assert(sample_args):
    args = sample_args
    xml_content = """<root>
    <item most-plausible-alternative="1" asks-for="unknown">
        <p>It started to rain.</p>
        <a1>I opened my umbrella.</a1>
        <a2>I wore sunglasses.</a2>
    </item>
    </root>"""
    with pytest.raises(AssertionError):
        with patch("xml.etree.ElementTree.parse") as mock_parse:
            mock_tree = ET.ElementTree(ET.fromstring(xml_content))
            mock_parse.return_value = mock_tree
            examples = copa_loader("dummy_path.xml", args)


def test_cqa_loader(sample_args):
    args = sample_args
    args.multiple_choice_prompt = "Answer the following question:"
    # Adjust the stem to end with a period to match the processing in cqa_loader
    json_line = json.dumps(
        {
            "answerKey": "A",
            "question": {
                "stem": "What is the capital of France.",
                "choices": [
                    {"text": "Paris"},
                    {"text": "London"},
                    {"text": "Berlin"},
                    {"text": "Madrid"},
                    {"text": "Rome"},
                ],
            },
        }
    )
    with patch("builtins.open", mock.mock_open(read_data=json_line)):
        examples = cqa_loader("dummy_path.jsonl", args)
        assert len(examples) == 1
        assert examples[0]["label"] == 0
        assert (
            "Answer the following question: Question:  What is the capital of France?"
            in examples[0]["premise"]
        )


def test_obqa_loader(sample_args):
    args = sample_args
    args.multiple_choice_prompt = "Answer the following question:"
    # Adjust the stem to end with a period to match the processing in cqa_loader
    json_line = json.dumps(
        {
            "answerKey": "A",
            "question": {
                "stem": "What is the capital of France.",
                "choices": [
                    {"text": "Paris", "label": 0},
                    {"text": "London", "label": 0},
                    {"text": "Berlin", "label": 1},
                    {"text": "Madrid", "label": 0},
                    {"text": "Rome", "label": 0},
                ],
            },
        }
    )
    with patch("builtins.open", mock.mock_open(read_data=json_line)):
        examples = obqa_loader("dummy_path.jsonl", args)
        assert len(examples) == 1
        assert examples[0]["label"] == 0
        assert (
            "Answer the following question: Question: What is the capital of France"
            in examples[0]["premise"]
        )


def test_generate_n_shot_demonstrations(sample_args):
    n_shot_dataset = [
        {
            "premise": "Question 1",
            "label": torch.tensor(0),
            "hypothesis0": "A1",
            "hypothesis1": "B1",
        },
        {
            "premise": "Question 2",
            "label": torch.tensor(1),
            "hypothesis0": "A2",
            "hypothesis1": "B2",
        },
    ]
    output = generate_n_shot_demonstrations(n_shot_dataset)
    expected_output = "Question 1A1\n\nQuestion 2B2\n\n"
    assert output == expected_output


def test_create_n_shot_splits(sample_args):
    args = sample_args
    args.n_shot = 1
    raw_dataset = MagicMock()
    n_shot_dataset = MagicMock()
    n_shot_dataset.shuffle.return_value.select.return_value = n_shot_dataset
    raw_dataset.shuffle.return_value.select.return_value = raw_dataset
    raw_dataset.map.return_value = raw_dataset
    # Adjust the patch path to match the module structure
    with patch(
        "mm_poe.methods.utils.data.generate_n_shot_demonstrations",
        return_value="Demo",
    ) as mock_generate:
        output_dataset, output_n_shot_dataset, n_shot_demonstrations = (
            create_n_shot_splits(raw_dataset, n_shot_dataset, args)
        )
        assert n_shot_demonstrations == "Demo"
        raw_dataset.map.assert_called_once()


def test_single_inference_loader(sample_args):
    args = sample_args
    path = "path/to/image.jpg"
    examples = single_inference_loader(path, args)
    assert len(examples) == 1
    assert examples[0]["image_path"] == path
    assert examples[0]["premise"].startswith(args.multiple_choice_prompt)


def test_anli_loader(sample_args):
    args = sample_args
    args.multiple_choice_prompt = None
    json_line = json.dumps(
        {
            "context": "A man is playing a piano.",
            "hypothesis": "The man is playing a musical instrument.",
            "label": "e",
        }
    )
    with patch("builtins.open", mock.mock_open(read_data=json_line)):
        examples = anli_loader(["dummy_path.jsonl"], args)
        assert len(examples) == 1
        assert examples[0]["label"] == 0
        assert (
            "A man is playing a piano. The man is playing a musical instrument."
            in examples[0]["premise"]
        )


def test_generate_n_shot_poe_demonstrations(sample_args):
    n_shot_dataset = [
        {
            "premise": "Question 1",
            "label": torch.tensor(0),
            "hypothesis0": "A1",
            "hypothesis1": "B1",
        },
        {
            "premise": "Question 2",
            "label": torch.tensor(1),
            "hypothesis0": "A2",
            "hypothesis1": "B2",
        },
    ]
    num_of_options = 2
    output, poe_output = generate_n_shot_poe_demonstrations(
        n_shot_dataset, num_of_options
    )
    assert isinstance(output, str)
    assert isinstance(poe_output, str)


def test_preprocess_function_seq2seq_channel(sample_args):
    examples = {
        "question": ["What is 2+2?"],
        "choice0": ["3"],
        "choice1": ["4"],
    }
    tokenizer = MagicMock()
    tokenizer.return_value = {
        "input_ids": [[1, 2], [3, 4]],
        "attention_mask": [[1, 1], [1, 1]],
    }
    kwargs = {
        "ending_names": ["choice0", "choice1"],
        "header_name": "question",
        "tokenizer": tokenizer,
    }
    output = preprocess_function_seq2seq_channel(examples, **kwargs)
    assert "header_input_ids" in output
    assert "ending_input_ids" in output


def test_preprocess_function_causal_channel(sample_args):
    examples = {
        "question": ["What is 2+2?"],
        "choice0": ["3"],
        "choice1": ["4"],
    }
    tokenizer = MagicMock()
    tokenizer.pad_token_id = 0
    # Adjust the tokenizer to return lists of lists
    tokenizer.return_value = {
        "input_ids": [[1, 2, 3], [4, 5, 6]],
        "attention_mask": [[1, 1, 1], [1, 1, 1]],
    }
    kwargs = {
        "ending_names": ["choice0", "choice1"],
        "header_name": "question",
        "tokenizer": tokenizer,
    }
    output = preprocess_function_causal_channel(examples, **kwargs)
    assert "input_ids" in output
    assert "labels" in output


def test_vqa_loader(sample_args):
    args = sample_args
    args.num_options = 2
    args.split = "val"
    ann_content = {
        "annotations": [{"multiple_choice_answer": "cat", "image_id": 123}]
    }
    ques_content = {
        "questions": [
            {
                "question": "What animal is this?",
                "multiple_choices": ["cat", "dog"],
            }
        ]
    }
    with patch("json.load", side_effect=[ann_content, ques_content]):
        with patch("os.path.join", return_value="path/to/image.jpg"):
            with patch("builtins.open", mock.mock_open()) as mock_file:
                # Mock the open calls for annotation and question files
                mock_file.side_effect = [
                    mock.mock_open(
                        read_data=json.dumps(ann_content)
                    ).return_value,
                    mock.mock_open(
                        read_data=json.dumps(ques_content)
                    ).return_value,
                ]
                examples = vqa_loader("dummy_path", args)
                assert len(examples) == 1
                assert examples[0]["label"] == 0
                assert examples[0]["image_path"] == "path/to/image.jpg"


def test_scienceqa_loader(sample_args):
    args = sample_args
    args.num_options = 4
    args.split = "val"
    ann_content = {
        "1": {
            "question": "What is H2O?",
            "choices": ["Water", "Oxygen", "Hydrogen", "Helium"],
            "answer": "0",
            "image": "image1.jpg",
        }
    }
    with patch("json.load", return_value=ann_content):
        with patch("os.listdir", return_value=["1"]):
            with patch("os.path.join", return_value="path/to/image.jpg"):
                with patch(
                    "builtins.open",
                    mock.mock_open(read_data=json.dumps(ann_content)),
                ):
                    examples = scienceqa_loader("dummy_path", args)
                    assert len(examples) == 1
                    assert examples[0]["label"] == 0
                    assert examples[0]["image_path"] == "path/to/image.jpg"


def test_ai2d_loader(sample_args):
    args = sample_args
    args.num_options = 3
    question_content = {
        "questions": {
            "What is this?": {
                "answerTexts": ["Cat", "Dog", "Mouse"],
                "correctAnswer": "1",
                "abcLabel": False,
            }
        },
        "imageName": "image1.jpg",
    }
    with patch("os.listdir", return_value=["file1.json"]):
        with patch("json.load", return_value=question_content):
            with patch(
                "os.path.join",
                side_effect=lambda *args: "path/to/" + "/".join(args[-2:]),
            ):
                with patch(
                    "builtins.open",
                    mock.mock_open(read_data=json.dumps(question_content)),
                ):
                    examples = ai2d_loader("dummy_path", args)
                    assert len(examples) == 1
                    assert examples[0]["label"] == 1
                    assert (
                        examples[0]["image_path"]
                        == "path/to/dummy_path/ai2d/images/image1.jpg"
                    )


# New test functions to increase coverage


def test_preprocess_function_causal_vqa(sample_args):
    examples = {
        "question": ["What is shown in the image?"],
        "choice0": ["Cat"],
        "choice1": ["Dog"],
        "image_path": ["path/to/image1.jpg"],
    }
    processor = MagicMock()
    tokenizer = MagicMock()
    tokenizer.pad_token_id = 0
    tokenizer.padding_side = "right"
    # Adjust the tokenizer to return lists of lists
    tokenizer.return_value = {
        "input_ids": [[1, 2], [3, 4]],
        "attention_mask": [[1, 1], [1, 1]],
    }
    processor.tokenizer = tokenizer
    data_obj = MagicMock()
    data_obj.data = {
        "pixel_values": torch.tensor(
            [[[1, 2], [3, 4]]] * 2
        )  # Repeat to match the number of choices
    }
    processor.image_processor.return_value = data_obj
    kwargs = {
        "ending_names": ["choice0", "choice1"],
        "header_name": "question",
        "image_header_name": "image_path",
        "processor": processor,
        "image_token": "",
    }
    with patch("PIL.Image.open", return_value=MagicMock(spec=Image.Image)):
        output = preprocess_function_causal_vqa(examples, **kwargs)
        assert "input_ids" in output
        assert "labels" in output
        assert "header_attention_mask" in output
        assert "ending_attention_mask" in output
        assert "images" in output


def test_preprocess_function_seq2seq_vqa_channel(sample_args):
    examples = {
        "question": ["What is shown in the image?"],
        "choice0": ["Cat"],
        "choice1": ["Dog"],
        "image_path": ["path/to/image1.jpg"],
    }
    processor = MagicMock()
    tokenizer = MagicMock()
    tokenizer.return_value = {
        "input_ids": [[1, 2], [3, 4]],
        "attention_mask": [[1, 1], [1, 1]],
    }
    processor.tokenizer = tokenizer
    data_obj = MagicMock()
    data_obj.data = {
        "pixel_values": torch.tensor(
            [[[1, 2], [3, 4]]] * 2
        )  # Repeat to match the number of choices
    }
    processor.image_processor.return_value = data_obj
    kwargs = {
        "ending_names": ["choice0", "choice1"],
        "header_name": "question",
        "image_header_name": "image_path",
        "processor": processor,
        "image_token": "",
    }
    with patch("PIL.Image.open", return_value=MagicMock(spec=Image.Image)):
        output = preprocess_function_seq2seq_vqa_channel(examples, **kwargs)
        assert "header_input_ids" in output
        assert "ending_input_ids" in output
        assert "images" in output
        assert len(output["images"]) == len(examples["question"])
        for img_list in output["images"]:
            assert len(img_list) == len(kwargs["ending_names"])


def test_preprocess_function_causal_vqa_channel(sample_args):
    examples = {
        "question": ["What is shown in the image?"],
        "hypothesis0": ["Cat"],
        "hypothesis1": ["Dog"],
        "image_path": ["path/to/image1.jpg"],
    }
    processor = MagicMock()
    tokenizer = MagicMock()
    tokenizer.pad_token_id = 0
    tokenizer.padding_side = "right"
    tokenizer.return_value = {
        "input_ids": [[1, 2], [3, 4]],
        "attention_mask": [[1, 1], [1, 1]],
    }
    processor.tokenizer = tokenizer
    data_obj = MagicMock()
    data_obj.data = {"pixel_values": torch.tensor([[[1, 2], [3, 4]]] * 2)}
    processor.image_processor.return_value = data_obj
    kwargs = {
        "ending_names": ["hypothesis0", "hypothesis1"],
        "header_name": "question",
        "image_header_name": "image_path",
        "processor": processor,
        "image_token": "",
    }
    with patch("PIL.Image.open", return_value=MagicMock(spec=Image.Image)):
        output = preprocess_function_causal_vqa_channel(examples, **kwargs)
        assert "input_ids" in output
        assert "labels" in output
        assert "header_attention_mask" in output
        assert "ending_attention_mask" in output
        assert "images" in output


def test_piqa_loader(sample_args):
    args = sample_args
    args.multiple_choice_prompt = "Answer the following question:"
    qa_content = json.dumps(
        {
            "goal": "To open a jar, you should",
            "sol1": "Twist the lid counter-clockwise",
            "sol2": "Push the lid upwards",
        }
    )
    label_content = "0\n"  # First solution is correct
    with patch("builtins.open", mock.mock_open()) as mock_file:
        mock_file.side_effect = [
            mock.mock_open(read_data=qa_content).return_value,
            mock.mock_open(read_data=label_content).return_value,
        ]
        examples = piqa_loader(
            ["dummy_qa_path.jsonl", "dummy_label_path.txt"], args
        )
        assert len(examples) == 1
        assert examples[0]["label"] == 0
        assert (
            "Answer the following question: Question: To open a jar, you should"
            in examples[0]["premise"]
        )


def test_qasc_loader(sample_args):
    args = sample_args
    args.multiple_choice_prompt = "Answer the following question:"
    json_line = json.dumps(
        {
            "answerKey": "B",
            "question": {
                "stem": "What do plants need to perform photosynthesis?",
                "choices": [
                    {"label": "A", "text": "Oxygen"},
                    {"label": "B", "text": "Sunlight"},
                    {"label": "C", "text": "Nitrogen"},
                    {"label": "D", "text": "Carbon dioxide"},
                    {"label": "E", "text": "Water"},
                    {"label": "F", "text": "Soil"},
                    {"label": "G", "text": "Minerals"},
                    {"label": "H", "text": "Glucose"},
                ],
            },
        }
    )
    with patch("builtins.open", mock.mock_open(read_data=json_line)):
        examples = qasc_loader("dummy_path.jsonl", args)
        assert len(examples) == 1
        assert examples[0]["label"] == 1  # 'B' corresponds to index 1
        assert (
            "Answer the following question: Question: What do plants need to perform photosynthesis?"
            in examples[0]["premise"]
        )


def test_siqa_loader(sample_args):
    args = sample_args
    args.multiple_choice_prompt = "Answer the following question:"
    qa_content = json.dumps(
        {
            "context": "Alex went to the store.",
            "question": "Why did Alex go to the store?",
            "answerA": "To buy groceries",
            "answerB": "To sell groceries",
            "answerC": "To sleep",
        }
    )
    label_content = "1\n"  # Answer index is 1 (but labels are 1-based in siqa_loader, and subtract 1)
    with patch("builtins.open", mock.mock_open()) as mock_file:
        mock_file.side_effect = [
            mock.mock_open(read_data=qa_content).return_value,
            mock.mock_open(read_data=label_content).return_value,
        ]
        examples = siqa_loader(
            ["dummy_qa_path.jsonl", "dummy_label_path.txt"], args
        )
        assert len(examples) == 1
        assert (
            examples[0]["label"] == 0
        )  # '1' in label file corresponds to index 0
        assert (
            "Answer the following question: Question: Alex went to the store. Why did Alex go to the store?"
            in examples[0]["premise"]
        )


def test_winogrande_loader(sample_args):
    args = sample_args
    args.multiple_choice_prompt = "Answer the following question:"
    qa_content = json.dumps(
        {
            "sentence": "The trophy doesn't fit in the brown suitcase because it's too big.",
            "option1": "trophy",
            "option2": "suitcase",
        }
    )
    label_content = "1\n"  # Correct answer is option1 (labels are 1-based)
    with patch("builtins.open", mock.mock_open()) as mock_file:
        mock_file.side_effect = [
            mock.mock_open(read_data=qa_content).return_value,
            mock.mock_open(read_data=label_content).return_value,
        ]
        examples = winogrande_loader(
            ["dummy_qa_path.jsonl", "dummy_label_path.txt"], args
        )
        assert len(examples) == 1
        assert (
            examples[0]["label"] == 0
        )  # '1' in label file corresponds to index 0
        assert (
            "Answer the following question: Question: The trophy doesn't fit in the brown suitcase because it's too big."
            in examples[0]["premise"]
        )


def test_date_understanding_loader(sample_args):
    args = sample_args
    args.multiple_choice_prompt = "Answer the following question:"
    args.num_options = 2
    data_content = {
        "task_prefix": "",
        "examples": [
            {"input": "What is 2+2?", "target_scores": {"4": 1, "5": 0}}
        ],
    }
    with patch("json.load", return_value=data_content):
        with patch(
            "builtins.open", mock.mock_open(read_data=json.dumps(data_content))
        ):
            examples = date_understanding_loader(["dummy_path.json"], args)
            assert len(examples) == 1
            assert examples[0]["label"] == 0  # '4' is at index 0
            assert (
                "Answer the following question: Question: What is 2+2?"
                in examples[0]["premise"]
            )


# Now this test_data.py covers only 60% of all tests when I run pytest coverage. You need to add more tests to this file. DO NOT CHANGE ANYTHING WRITTEN IN THIS FILE, NO CHANGE TO ANY OF THE CURRENT TESTS, ALL OF THEM ARE WORKING. ADD NEW TESTS TO THIS FILE TO GET 100% COVERAGE



import argparse
import csv
import os
import random

import numpy as np
import torch

from transformers import (
    AutoTokenizer,
    AutoModelForCausalLM,
    AutoModelForSeq2SeqLM,
    AutoProcessor,
    AutoModelForVision2Seq,
    BitsAndBytesConfig,
)
from datasets import Dataset

# import data.py, which is located in the same directory

from .data import (
    copa_loader,
    cqa_loader,
    obqa_loader,
    piqa_loader,
    qasc_loader,
    siqa_loader,
    winogrande_loader,
    date_understanding_loader,
    anli_loader,
    vqa_loader,
    scienceqa_loader,
    ai2d_loader,
    single_inference_loader,
    custom_loader,
)


def set_seed(seed):
    os.environ["PYTHONHASHSEED"] = str(seed)
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.cuda.manual_seed_all(seed)


def parse_args():
    parser = argparse.ArgumentParser("Inference on multiple choice benchmarks")
    parser.add_argument(
        "--seed",
        type=int,
        default=0,
        help="Random seed for reproducibility.",
    )
    parser.add_argument(
        "--model_family",
        type=str,
        choices=[
            "GPT2",
            "T5",
            "FLAN-T5",
            "Pythia",
            "OPT-IML",
            "Dolly",
            "BLIP2-OPT",
            "BLIP2-T5",
            "InstructBLIP",
            "GIT",
            "PaliGemma",
            "Idefics2",
        ],
        default=None,
        required=True,
        help="The moddel family, as checkpoints under the same "
        + "model family use same codes to download.",
    )
    parser.add_argument(
        "--checkpoint",
        type=str,
        default=None,
        required=True,
        help="The checkpoint name under a model family, "
        + "e.g. gpt2, gpt2-medium, gpt2-large, gpt2-xl.",
    )
    parser.add_argument(
        "--amateur_checkpoint",
        type=str,
        default=None,
        help="The amateur checkpoint name under a model family. "
        + "For constrative decoding.",
    )
    parser.add_argument(
        "--expert_method",
        type=str,
        choices=[
            "language_modeling",
            "calibration",
            "channel",
            "multiple_choice_prompt",
        ],
        default="language_modeling",
        help="The expert method. For constrative decoding.",
    )
    parser.add_argument(
        "--amateur_method",
        type=str,
        choices=[
            "language_modeling",
            "calibration",
            "channel",
            "multiple_choice_prompt",
        ],
        default="language_modeling",
        help="The amateur method. For constrative decoding.",
    )
    parser.add_argument(
        "--weighting_parameter",
        type=float,
        default=-1.0,
        help="The weighting parameter for constrative decoding. "
        + "It is applied to the amateur model.",
    )
    parser.add_argument(
        "--weighting_parameters",
        type=str,
        default=None,
        help="The weighting parameters for constrative decoding. "
        + "One weight for one dataset.",
    )
    parser.add_argument(
        "--num_random_search",
        type=int,
        default=0,
        help="The number of random search for the "
        + "weighting parameter for constrative decoding.",
    )
    parser.add_argument(
        "--loading_precision",
        type=str,
        choices=["FP32", "FP16", "BF16", "INT8", "INT4"],
        default="FP32",
        help="The precision of the model to be loaded.",
    )
    parser.add_argument(
        "--datasets",
        type=str,
        # choices=["copa", "cqa", "winogrande"],
        default=None,
        required=True,
        help="The datasets to inference on. "
        + "Pass multiple datasets separate by space",
    )
    parser.add_argument(
        "--sample",
        type=int,
        default=None,
        help="The number of samples to inference on. "
        + "If None, inference on the whole dataset.",
    )
    parser.add_argument(
        "--batch_size",
        type=int,
        default=32,
        help="Batch size for inference.",
    )
    parser.add_argument(
        "--n_shot",
        type=int,
        default=0,
        help="Number of few-shot demonstrations. 0 means zero-shot.",
    )
    parser.add_argument(
        "--multiple_choice_prompt",
        type=str,
        default=None,
        help="The multiple choice prompt.",
    )
    parser.add_argument(
        "--calibration_prompt",
        type=str,
        default=None,
        help="Calibration prompt, as in P(y|x)/P(y|calibration_prompt).",
    )
    parser.add_argument(
        "--do_channel",
        action="store_true",
        help="Whether to do channel, i.e., P(x|y_i).",
    )
    parser.add_argument(
        "--process_of_elimination_prompt",
        type=str,
        default=None,
        help="The process of elimination prompt. "
        + "It asks the model to ignore masked options.",
    )
    parser.add_argument(
        "--scoring_method_for_process_of_elimination",
        type=str,
        choices=[
            "language_modeling",
            "calibration",
            "channel",
            "multiple_choice_prompt",
        ],
        default="language_modeling",
        help="The scoring method for process of elimination.",
    )
    parser.add_argument(
        "--prompting_method_for_process_of_elimination",
        type=str,
        choices=["multiple_choice_prompt"],
        default="multiple_choice_prompt",
        help="The prompting method for process of elimination.",
    )
    parser.add_argument(
        "--mask_strategy_for_process_of_elimination",
        type=str,
        choices=["lowest", "below_average", "min_k"],
        default="lowest",
        help="The mask strategy for process of elimination.",
    )
    parser.add_argument(
        "--do_synonym",
        action="store_true",
        help="Whether to generate synonyms for options.",
    )
    parser.add_argument(
        "--number_of_synonyms",
        type=int,
        default=5,
        help="The number of synonyms to be used in the generative method.",
    )
    parser.add_argument(
        "--generate_synonyms_prompt",
        type=str,
        default=None,
        help="The prompt template for generating synonyms. "
        + "'option is replaced with actual options'",
    )
    parser.add_argument(
        "--push_data_to_hub",
        action="store_true",
        help="Whether to push the data to Hugging Face Hub. "
        + "This is convienient for LLM experiments.",
    )
    parser.add_argument(
        "--min_k",
        type=int,
        help="A masking strategy that masks the k most unlikely options",
    )
    parser.add_argument(
        "--mask_token",
        type=str,
        default=None,
        help="The mask token. If None, "
        + "use the default mask token of the model.",
    )

    args = parser.parse_args()
    return args


def load_data(args):
    # load test data for final performance.
    # load dev data to tune hyperparameters.
    # commonsense reasoning datasets
    train_file_path = None
    if args.dataset == "copa":
        ending_names = ["hypothesis0", "hypothesis1"]
        header_name = "premise"
        file_path = os.path.join("../data", args.dataset, "copa-test.xml")
        train_file_path = os.path.join("../data", args.dataset, "copa-dev.xml")
        loader = copa_loader
    elif args.dataset == "cqa":
        ending_names = [
            "hypothesis0",
            "hypothesis1",
            "hypothesis2",
            "hypothesis3",
            "hypothesis4",
        ]
        header_name = "premise"
        file_path = os.path.join("../data", args.dataset, "dev.jsonl")
        train_file_path = os.path.join("../data", args.dataset, "train.jsonl")
        loader = cqa_loader
    elif args.dataset == "obqa":
        ending_names = [
            "hypothesis0",
            "hypothesis1",
            "hypothesis2",
            "hypothesis3",
        ]
        header_name = "premise"
        file_path = os.path.join("../data", args.dataset, "test.jsonl")
        train_file_path = os.path.join("../data", args.dataset, "train.jsonl")
        loader = obqa_loader
    elif args.dataset == "piqa":
        ending_names = ["hypothesis0", "hypothesis1"]
        header_name = "premise"
        data_path = os.path.join("../data", args.dataset, "valid.jsonl")
        label_path = os.path.join("../data", args.dataset, "valid-labels.lst")
        file_path = [data_path, label_path]
        train_file_path = [
            path.replace("valid", "train") for path in file_path
        ]
        loader = piqa_loader
    elif args.dataset == "qasc":
        ending_names = [
            "hypothesis0",
            "hypothesis1",
            "hypothesis2",
            "hypothesis3",
            "hypothesis4",
            "hypothesis5",
            "hypothesis6",
            "hypothesis7",
        ]
        header_name = "premise"
        file_path = os.path.join("../data", args.dataset, "dev.jsonl")
        train_file_path = os.path.join("../data", args.dataset, "train.jsonl")
        loader = qasc_loader
    elif args.dataset == "siqa":
        ending_names = ["hypothesis0", "hypothesis1", "hypothesis2"]
        header_name = "premise"
        data_path = os.path.join("../data", args.dataset, "dev.jsonl")
        label_path = os.path.join("../data", args.dataset, "dev-labels.lst")
        file_path = [data_path, label_path]
        train_file_path = [path.replace("dev", "train") for path in file_path]
        loader = siqa_loader
    elif args.dataset == "winogrande":
        ending_names = ["hypothesis0", "hypothesis1"]
        header_name = "premise"
        data_path = os.path.join("../data", args.dataset, "dev.jsonl")
        label_path = os.path.join("../data", args.dataset, "dev-labels.lst")
        file_path = [data_path, label_path]
        train_file_path = [
            path.replace("dev", "train_xs") for path in file_path
        ]
        loader = winogrande_loader
    # BIG-Bench tasks
    elif args.dataset == "disambiguation_qa":
        args.num_options = 3
        ending_names = [f"hypothesis{i}" for i in range(args.num_options)]
        header_name = "premise"
        file_path = [
            os.path.join("../data", "big_bench", f"{args.dataset}.json")
        ]
        loader = date_understanding_loader
    elif args.dataset == "conceptual_combinations":
        args.num_options = 4
        ending_names = [f"hypothesis{i}" for i in range(args.num_options)]
        header_name = "premise"
        file_path = []
        file_suffixes = [
            "contradictions",
            "emergent_properties",
            "fanciful_fictional_combinations",
            "homonyms",
            "invented_words",
            "surprising_uncommon_combinations",
        ]
        for suffix in file_suffixes:
            file_path.append(
                os.path.join(
                    "../data", "big_bench", f"{args.dataset}_{suffix}.json"
                )
            )
        loader = date_understanding_loader
    elif args.dataset == "date_understanding":
        args.num_options = 6
        ending_names = [f"hypothesis{i}" for i in range(args.num_options)]
        header_name = "premise"
        file_path = [
            os.path.join("../data", "big_bench", f"{args.dataset}.json")
        ]
        loader = date_understanding_loader
    elif args.dataset in [
        "emoji_movie",
        "evaluating_information_essentiality",
        "logical_args",
        "riddle_sense",
    ]:
        args.num_options = 5
        ending_names = [f"hypothesis{i}" for i in range(args.num_options)]
        header_name = "premise"
        file_path = [
            os.path.join("../data", "big_bench", f"{args.dataset}.json")
        ]
        loader = date_understanding_loader
    elif args.dataset in [
        "ruin_names",
        "temporal_sequences",
        "code_line_description",
        "crass_ai",
        "identify_math_theorems",
        "identify_odd_metaphor",
    ]:
        args.num_options = 4
        ending_names = [f"hypothesis{i}" for i in range(args.num_options)]
        header_name = "premise"
        file_path = [
            os.path.join("../data", "big_bench", f"{args.dataset}.json")
        ]
        loader = date_understanding_loader
    elif args.dataset == "penguins_in_a_table":
        # use the five object subtask
        args.num_options = 5
        ending_names = [f"hypothesis{i}" for i in range(args.num_options)]
        header_name = "premise"
        file_path = [
            os.path.join("../data", "big_bench", f"{args.dataset}.json")
        ]
        loader = date_understanding_loader
    elif args.dataset == "strange_stories":
        args.num_options = 4
        ending_names = [f"hypothesis{i}" for i in range(args.num_options)]
        header_name = "premise"
        file_path = [
            os.path.join(
                "../data", "big_bench", f"{args.dataset}_multiple_choice.json"
            )
        ]
        loader = date_understanding_loader
    elif args.dataset == "reasoning_about_colored_objects":
        args.num_options = 18
        ending_names = [f"hypothesis{i}" for i in range(args.num_options)]
        header_name = "premise"
        file_path = [
            os.path.join("../data", "big_bench", f"{args.dataset}.json")
        ]
        loader = date_understanding_loader
    elif args.dataset == "symbol_interpretation":
        args.num_options = 5
        ending_names = [f"hypothesis{i}" for i in range(args.num_options)]
        header_name = "premise"
        file_path = []
        file_suffixes = [
            "adversarial",
            "emoji_agnostic",
            "name_agnostic",
            "plain",
            "tricky",
        ]
        for suffix in file_suffixes:
            file_path.append(
                os.path.join(
                    "../data", "big_bench", f"{args.dataset}_{suffix}.json"
                )
            )
        loader = date_understanding_loader
    elif args.dataset == "tracking_shuffled_objects":
        # use the five object subtask
        args.num_options = 5
        ending_names = [f"hypothesis{i}" for i in range(args.num_options)]
        header_name = "premise"
        file_suffix = "five_objects"
        file_path = [
            os.path.join(
                "../data", "big_bench", f"{args.dataset}_{file_suffix}.json"
            )
        ]
        loader = date_understanding_loader
    elif args.dataset in [
        "logical_deduction_three_objects",
        "logical_deduction_five_objects",
        "logical_deduction_seven_objects",
    ]:
        if "three" in args.dataset:
            args.num_options = 3
        elif "five" in args.dataset:
            args.num_options = 5
        else:
            args.num_options = 7
        ending_names = [f"hypothesis{i}" for i in range(args.num_options)]
        header_name = "premise"
        file_path = [
            os.path.join("../data", "big_bench", f"{args.dataset}.json")
        ]
        loader = date_understanding_loader
    # other datasets
    elif args.dataset == "anli":
        ending_names = ["hypothesis0", "hypothesis1", "hypothesis2"]
        header_name = "premise"
        file_path = []
        file_prefixes = ["R1", "R2", "R3"]
        for prefix in file_prefixes:
            file_path.append(
                os.path.join(
                    "../data", f"{args.dataset}", f"{prefix}_dev.jsonl"
                )
            )
        train_file_path = [path.replace("dev", "train") for path in file_path]
        loader = anli_loader
    elif args.dataset in ["anli_r1", "anli_r2", "anli_r3"]:
        args.num_options = 3
        ending_names = [f"hypothesis{i}" for i in range(args.num_options)]
        # ending_names = ['hypothesis0', 'hypothesis1', 'hypothesis2']
        header_name = "premise"
        prefix = args.dataset.split("_")[-1]
        file_path = [
            os.path.join("../data", "anli", f"{prefix.capitalize()}_dev.jsonl")
        ]
        # file_path = []
        # # file_prefixes = ["R1", "R2", "R3"]
        # for prefix in file_prefixes:
        #     file_path.append(os.path.join("../data", f"{args.dataset}", \
        #                                   f"{prefix}_dev.jsonl"))
        train_file_path = [path.replace("dev", "train") for path in file_path]
        loader = anli_loader
    elif args.dataset == "vqa":
        args.num_options = 18
        args.split = "val"
        file_path = os.path.join("../data", args.dataset)
        train_file_path = os.path.join("../data", args.dataset)
        ending_names = [f"hypothesis{i}" for i in range(args.num_options)]
        header_name = "premise"
        image_header_name = "image_path"
        loader = vqa_loader
    elif args.dataset == "scienceqa":
        args.num_options = 4
        args.split = "val"
        file_path = os.path.join("../data", args.dataset)
        train_file_path = os.path.join("../data", args.dataset)
        ending_names = [f"hypothesis{i}" for i in range(args.num_options)]
        header_name = "premise"
        image_header_name = "image_path"
        loader = scienceqa_loader
    elif args.dataset == "ai2d":
        args.num_options = 4
        file_path = os.path.join("../data", args.dataset)
        ending_names = [f"hypothesis{i}" for i in range(args.num_options)]
        header_name = "premise"
        image_header_name = "image_path"
        loader = ai2d_loader
    elif args.dataset == "single_inference":
        file_path = args.image_path
        ending_names = [f"hypothesis{i}" for i in range(args.num_options)]
        header_name = "premise"
        image_header_name = "image_path"
        loader = single_inference_loader
    elif args.dataset == "custom_dataset":
        args.num_options = 4
        args.split = "val"
        file_path = os.path.join("../data", args.dataset)
        train_file_path = os.path.join("../data", args.dataset)
        ending_names = [f"hypothesis{i}" for i in range(args.num_options)]
        header_name = "premise"
        image_header_name = "image_path"
        loader = custom_loader
    else:
        print(f"{args.dataset}: downloader not implemented.")
        return

    dev_data = loader(file_path, args)
    dev_dataset = Dataset.from_list(dev_data).with_format("torch")
    if train_file_path is not None:
        args.split = "train"
        train_data = loader(train_file_path, args)
        train_dataset = Dataset.from_list(train_data).with_format("torch")
    else:  # BB tasks have no train set.
        train_dataset = dev_dataset
    if args.dataset in [
        "vqa",
        "scienceqa",
        "ai2d",
        "single_inference",
        "custom_dataset",
    ]:
        return (
            ending_names,
            header_name,
            image_header_name,
            dev_dataset,
            train_dataset,
        )
    return ending_names, header_name, dev_dataset, train_dataset


def load_model(device, model_path, args):
    if args.model_family in ["GPT2", "Pythia", "OPT-IML", "Dolly"]:
        tokenizer_func = AutoTokenizer
        model_func = AutoModelForCausalLM
    elif args.model_family in ["T5", "FLAN-T5"]:
        tokenizer_func = AutoTokenizer
        model_func = AutoModelForSeq2SeqLM
    elif args.model_family in [
        "BLIP2-OPT",
        "BLIP2-T5",
        "InstructBLIP",
        "GIT",
        "PaliGemma",
        "Idefics2",
    ]:
        tokenizer_func = AutoProcessor
        model_func = AutoModelForVision2Seq
    else:
        print(f"{args.model_family}: downloader not implemented.")
        return
    if args.model_family == "Dolly":
        tokenizer = tokenizer_func.from_pretrained(
            model_path, padding_side="left"
        )
    elif args.model_family == "Idefics2":
        tokenizer = tokenizer_func.from_pretrained(
            model_path, do_image_splitting=False
        )
    else:
        tokenizer = tokenizer_func.from_pretrained(model_path)
    if args.model_family in ["GPT2", "Pythia", "Dolly"]:
        tokenizer.pad_token = tokenizer.eos_token
    elif args.model_family in ["PaliGemma", "GIT"]:
        tokenizer.tokenizer.padding_side = "left"

    # load with different precision
    if args.loading_precision == "FP16":
        model = model_func.from_pretrained(
            model_path, device_map=device, torch_dtype=torch.float16
        )
    elif args.loading_precision == "BF16":
        model = model_func.from_pretrained(
            model_path, device_map=device, torch_dtype=torch.bfloat16
        )
    elif args.loading_precision == "INT8":
        quantization_config = BitsAndBytesConfig(
            load_in_8bit=True, llm_int8_threshold=200.0
        )
        model = model_func.from_pretrained(
            model_path,
            torch_dtype=torch.float16,
            device_map=device,
            quantization_config=quantization_config,
        )
    elif args.loading_precision == "INT4":
        quantization_config = BitsAndBytesConfig(
            load_in_4bit=True,
            bnb_4bit_quant_type="nf4",
            bnb_4bit_use_double_quant=True,
            bnb_4bit_compute_dtype=torch.bfloat16,
        )
        model = model_func.from_pretrained(
            model_path,
            device_map=device,
            quantization_config=quantization_config,
        )
    else:  # FP32
        model = model_func.from_pretrained(model_path, device_map=device)
    print(
        f"Memory footprint: {model.get_memory_footprint() / 1024 **3:.2f} GB."
    )
    return model, tokenizer


def write_to_csv(save_path, args, total_accuracy):
    os.makedirs(os.path.dirname(save_path), exist_ok=True)
    csv_exists = os.path.isfile(save_path)
    with open(save_path, "a+", newline="") as csvfile:
        csvwriter = csv.writer(csvfile)
        if args.method == "process_of_elimination":
            if not csv_exists:
                csvwriter.writerow(
                    [
                        "model_family",
                        "checkpoint",
                        "loading_precision",
                        "dataset",
                        "batch_size",
                        "method",
                        "scoring_method",
                        "prompting_method",
                        "mask_strategy",
                        "mask_token",
                        "seed",
                        "n_shot",
                        "sample",
                        "mask_accuracy",
                        "accuracy",
                    ]
                )
            csvwriter.writerow(
                [
                    args.model_family,
                    args.checkpoint,
                    args.loading_precision,
                    args.dataset,
                    args.batch_size,
                    args.method,
                    args.scoring_method_for_process_of_elimination,
                    args.prompting_method_for_process_of_elimination,
                    args.mask_strategy_for_process_of_elimination,
                    args.mask_token,
                    args.seed,
                    int(args.n_shot),
                    args.sample,
                    f"{args.mask_accuracy:.4f}",
                    f"{total_accuracy:.4f}",
                ]
            )
        elif args.method == "contrastive_decoding":
            if not csv_exists:
                csvwriter.writerow(
                    [
                        "model_family",
                        "checkpoint",
                        "amateur_checkpoint",
                        "loading_precision",
                        "dataset",
                        "batch_size",
                        "method",
                        "expert_method",
                        "amateur_method",
                        "weighting_parameter",
                        "seed",
                        "n_shot",
                        "sample",
                        "expert_accuracy",
                        "amateur_accuracy",
                        "accuracy",
                    ]
                )
            csvwriter.writerow(
                [
                    args.model_family,
                    args.checkpoint,
                    args.amateur_checkpoint,
                    args.loading_precision,
                    args.dataset,
                    args.batch_size,
                    args.method,
                    args.expert_method,
                    args.amateur_method,
                    args.weighting_parameter,
                    args.seed,
                    int(args.n_shot),
                    args.sample,
                    f"{args.expert_accuracy:.4f}",
                    f"{args.amateur_accuracy:.4f}",
                    f"{total_accuracy:.4f}",
                ]
            )
        elif args.method == "generate_synonyms":
            if not csv_exists:
                csvwriter.writerow(
                    [
                        "model_family",
                        "checkpoint",
                        "loading_precision",
                        "dataset",
                        "batch_size",
                        "method",
                        "number_of_synonyms",
                        "seed",
                        "n_shot",
                        "sample",
                        "accuracy",
                    ]
                )
            csvwriter.writerow(
                [
                    args.model_family,
                    args.checkpoint,
                    args.loading_precision,
                    args.dataset,
                    args.batch_size,
                    args.method,
                    args.number_of_synonyms,
                    args.seed,
                    int(args.n_shot),
                    args.sample,
                    f"{total_accuracy:.4f}",
                ]
            )
        else:
            if not csv_exists:
                csvwriter.writerow(
                    [
                        "model_family",
                        "checkpoint",
                        "loading_precision",
                        "dataset",
                        "batch_size",
                        "method",
                        "seed",
                        "n_shot",
                        "sample",
                        "accuracy",
                    ]
                )
            csvwriter.writerow(
                [
                    args.model_family,
                    args.checkpoint,
                    args.loading_precision,
                    args.dataset,
                    args.batch_size,
                    args.method,
                    args.seed,
                    int(args.n_shot),
                    args.sample,
                    f"{total_accuracy:.4f}",
                ]
            )
